// © 2022 Daily, Co. All Rights Reserved

import 'package:freezed_annotation/freezed_annotation.dart';

part 'devices.freezed.dart';
part 'devices.g.dart';

@Freezed(toJson: false, fromJson: false)
class DeviceId with _$DeviceId {
  const factory DeviceId(String id) = _DeviceId;

  const DeviceId._();

  factory DeviceId.fromJson(String value) => DeviceId(value);

  static const speakerPhone = DeviceId('SPEAKERPHONE');
  static const wired = DeviceId('WIRED');
  static const earpiece = DeviceId('EARPIECE');
  static const bluetooth = DeviceId('BLUETOOTH');

  String toJson() => id;
}

@Freezed(toJson: false, fromJson: false)
class DeviceGroupId with _$DeviceGroupId {
  const factory DeviceGroupId(String id) = _DeviceGroupId;

  const DeviceGroupId._();

  factory DeviceGroupId.fromJson(String value) => DeviceGroupId(value);

  String toJson() => id;
}

/// Describes available media input and output devices.
@freezed
class Devices with _$Devices {
  const factory Devices({
    required List<Device> camera,
    required List<Device> microphone,
    required List<Device> speaker,
  }) = _Devices;

  const Devices._();

  factory Devices.empty() => const Devices(camera: [], microphone: [], speaker: []);

  factory Devices.fromJson(Map<String, dynamic> json) => _$DevicesFromJson(json);

  factory Devices.fromIterable(Iterable<Device> devices) {
    final camera = <Device>[];
    final microphone = <Device>[];
    final speaker = <Device>[];
    for (final device in devices) {
      switch (device.kind) {
        case DeviceKind.videoInput:
          camera.add(device);
          break;
        case DeviceKind.audioInput:
          microphone.add(device);
          break;
        case DeviceKind.audioOutput:
          speaker.add(device);
          break;
      }
    }
    return Devices(
      camera: List.unmodifiable(camera),
      microphone: List.unmodifiable(microphone),
      speaker: List.unmodifiable(speaker),
    );
  }

  Set<DeviceId> get ids => (camera + microphone + speaker).map((device) => device.deviceId).toSet();
}

Map<String, DeviceKind> _jsonValues = {for (final action in DeviceKind.values) action.value: action};

enum DeviceKind {
  videoInput('videoinput'),
  audioInput('audioinput'),
  audioOutput('audiooutput');

  const DeviceKind(this.value);

  factory DeviceKind.fromJson(String value) =>
      _jsonValues[value] ?? (throw ArgumentError('Unsupported CallState: "$value"'));

  final String value;
}

/// Describes a single media input or output device.
@freezed
class Device with _$Device {
  const factory Device({
    /// Identifier for the represented device that is persistent across application launches.
    required DeviceId deviceId,
    required DeviceGroupId groupId,

    /// The type of this device.
    required DeviceKind kind,

    /// A label describing the device, eg. "External USB Webcam"
    required String label,
  }) = _Device;

  factory Device.fromJson(Map<String, dynamic> json) => _$DeviceFromJson(json);
}
