// © 2022 Daily, Co. All Rights Reserved

import 'package:freezed_annotation/freezed_annotation.dart';

import '../util.dart';
import 'common.dart';

part 'participants.freezed.dart';
part 'participants.g.dart';

@Freezed(copyWith: false)
class ParticipantCounts with _$ParticipantCounts {
  const factory ParticipantCounts({required int present, required int hidden}) = _ParticipantCounts;
  factory ParticipantCounts.fromJson(Map<String, dynamic> json) => _$ParticipantCountsFromJson(json);
}

Map<String, dynamic> _participantMapToJson(Map<ParticipantId, Participant> value) =>
    value.map((key, value) => MapEntry(key.toJson(), value.toJson()));

Map<ParticipantId, Participant> _participantMapFromJson(Map<String, dynamic> obj) =>
    obj.map((key, value) => MapEntry(ParticipantId.fromJson(key), Participant.fromJson(value)));

@Freezed(fromJson: false)
class Participants with _$Participants {
  const factory Participants({
    required Participant local,
    @Default({})
    // ignore: invalid_annotation_target
    @JsonKey(toJson: _participantMapToJson, fromJson: _participantMapFromJson)
        Map<ParticipantId, Participant> remote,
    @Default({})
    // ignore: invalid_annotation_target
    @JsonKey(toJson: _participantMapToJson, fromJson: _participantMapFromJson)
        Map<ParticipantId, Participant> all,
  }) = _Participants;

  const Participants._();

  factory Participants.fromJson(Map<String, dynamic> json) {
    final remoteParticipants = Map.fromEntries(
      json.entries.expand<MapEntry<ParticipantId, Participant>>((entry) sync* {
        if (entry.key != 'local') {
          final participant = Participant.fromJson(entry.value);
          if (!participant.info.isLocal) {
            yield MapEntry(participant.id, participant);
          }
        }
      }),
    );
    // ignore: avoid_as
    final localParticipant = (json['local'] as Map<String, dynamic>?)?.let(Participant.fromJson);
    if (localParticipant == null) throw ArgumentError('Could not find local participant');
    return Participants(
      local: localParticipant,
      remote: Map.unmodifiable(remoteParticipants),
      all: Map.unmodifiable({localParticipant.id: localParticipant, ...remoteParticipants}),
    );
  }

  Participants upsertParticipant(Participant updatedParticipant) => copyWith(
        local: updatedParticipant.info.isLocal ? updatedParticipant : local,
        remote: updatedParticipant.info.isLocal
            ? remote
            : Map.unmodifiable({...remote, updatedParticipant.id: updatedParticipant}),
        all: Map.unmodifiable({...all, updatedParticipant.id: updatedParticipant}),
      );

  Participants removeParticipant(Participant participant) => copyWith(
        remote: Map.unmodifiable({...remote}..remove(participant.id)),
        all: Map.unmodifiable({...all}..remove(participant.id)),
      );
}

@freezed
class Participant with _$Participant {
  const factory Participant({
    required ParticipantId id,
    required ParticipantInfo info,
    ParticipantMedia? media,
  }) = _Participant;
  const Participant._();

  factory Participant.fromJson(Map<String, dynamic> json) => _$ParticipantFromJson(json);

  bool get isCameraMuted {
    final camState = media?.camera.state;
    return camState == null || camState == MediaState.blocked || camState == MediaState.off;
  }

  bool get isMicrophoneMuted {
    final micState = media?.microphone.state;
    return micState == null || micState == MediaState.blocked || micState == MediaState.off;
  }
}

@freezed
class ParticipantMedia with _$ParticipantMedia {
  const factory ParticipantMedia({
    required ParticipantVideoInfo camera,
    required ParticipantAudioInfo microphone,
    required ParticipantVideoInfo screenVideo,
    required ParticipantAudioInfo screenAudio,
  }) = _ParticipantMedia;

  factory ParticipantMedia.fromJson(Map<String, dynamic> json) => _$ParticipantMediaFromJson(json);
}

@freezed
class MediaStreamTrack with _$MediaStreamTrack {
  const factory MediaStreamTrack({required String id}) = _MediaStreamTrack;

  factory MediaStreamTrack.fromJson(Map<String, dynamic> json) => _$MediaStreamTrackFromJson(json);
}

@freezed
class ParticipantVideoInfo with _$ParticipantVideoInfo {
  const factory ParticipantVideoInfo({
    required MediaState state,
    required TrackSubscriptionState subscribed,
    required MediaStreamTrack? track,
  }) = _ParticipantVideoInfo;

  factory ParticipantVideoInfo.fromJson(Map<String, dynamic> json) => _$ParticipantVideoInfoFromJson(json);
}

@freezed
class ParticipantAudioInfo with _$ParticipantAudioInfo {
  const factory ParticipantAudioInfo({
    required MediaState state,
    required TrackSubscriptionState subscribed,
  }) = _ParticipantAudioInfo;

  factory ParticipantAudioInfo.fromJson(Map<String, dynamic> json) => _$ParticipantAudioInfoFromJson(json);
}

/// - `blocked`: The track is blocked, i.e. does not have permissions.
/// - `off`: The track is off, but not blocked, ie. `isEnabled` or `isPublishing` is `false`,
/// - `receivable`: The track has been published and is available to be received, but hasn't been subscribed to.
/// - `loading`: The track is loading. It has been subscribed to.
/// - `playable`: The track is ready to be played. It has been subscribed to and has finished loading.
/// - `interrupted`: The track is unexpectedly (and hopefully only temporarily) unplayable.
enum MediaState { blocked, off, receivable, loading, playable, interrupted }

enum TrackSubscriptionState { subscribed, staged, unsubscribed }

@Freezed(toJson: false, fromJson: false)
class ParticipantId with _$ParticipantId implements Comparable<ParticipantId> {
  const factory ParticipantId(String id) = _ParticipantId;
  const ParticipantId._();

  factory ParticipantId.fromJson(String value) => ParticipantId(value);
  String toJson() => id;

  @override
  int compareTo(ParticipantId other) => id.compareTo(other.id);
}

@freezed
class ParticipantInfo with _$ParticipantInfo {
  factory ParticipantInfo({
    // ignore: invalid_annotation_target
    @JsonKey(fromJson: dateTimeFromJson, toJson: dateTimeToJson) DateTime? joinedAt,
    required bool isLocal,
    required bool isOwner,
    ParticipantId? userId,
    // ignore: invalid_annotation_target
    @JsonKey(name: 'userName') String? username,
  }) = _ParticipantInfo;

  factory ParticipantInfo.fromJson(Map<String, dynamic> json) => _$ParticipantInfoFromJson(json);
}
