// © 2022 Daily, Co. All Rights Reserved

// ignore_for_file: invalid_annotation_target

import 'package:freezed_annotation/freezed_annotation.dart';

import '../../daily_flutter.dart';

part 'streaming.freezed.dart';
part 'streaming.g.dart';

enum StreamingStartMode {
  /// Always create a new stream, even if there is already one in progress.
  forceNew,

  /// Only create a new stream if one does not already exist.
  onlyIfNotAlreadyStreaming,
}

enum PortraitVariant {
  /// A vertical grid layout -- this is the default.
  vertical,

  /// Use a large view for one participant, and a smaller rectangle for the other participant. Participants' videos are
  /// cropped and scaled to fill the entire available space.
  inset
}

@Freezed(toJson: false, fromJson: false)
class StreamId with _$StreamId {
  const factory StreamId(String id) = _StreamId;

  const StreamId._();

  factory StreamId.fromJson(String value) => StreamId(value);

  String toJson() => id;
}

@Freezed(toJson: false, fromJson: false)
class RecordingId with _$RecordingId {
  const factory RecordingId(String id) = _RecordingId;

  const RecordingId._();

  factory RecordingId.fromJson(String value) => RecordingId(value);

  String toJson() => id;
}

@freezed
class StreamingSettings with _$StreamingSettings {
  const factory StreamingSettings({
    StreamingVideoSettings? video,
    StreamingAudioSettings? audio,
    StreamingLayout? layout,
    int? maxDuration,
  }) = _StreamingSettings;

  factory StreamingSettings.fromJson(Map<String, dynamic> json) => _$StreamingSettingsFromJson(json);
}

@Freezed(toJson: true)
class StreamingSettingsUpdate with _$StreamingSettingsUpdate {
  const factory StreamingSettingsUpdate({
    StreamingLayout? layout,
  }) = _StreamingSettingsUpdate;
}

@freezed
class StreamingVideoSettings with _$StreamingVideoSettings {
  const factory StreamingVideoSettings({
    int? width,
    int? height,
    int? fps,
    int? bitrate,
    String? backgroundColor,
  }) = _StreamingVideoSettings;

  factory StreamingVideoSettings.fromJson(Map<String, dynamic> json) => _$StreamingVideoSettingsFromJson(json);
}

@freezed
class StreamingAudioSettings with _$StreamingAudioSettings {
  const factory StreamingAudioSettings({int? bitrate}) = _StreamingAudioSettings;

  factory StreamingAudioSettings.fromJson(Map<String, dynamic> json) => _$StreamingAudioSettingsFromJson(json);
}

@Freezed(unionKey: 'preset')
class StreamingLayout with _$StreamingLayout {
  /// The default grid layout, which renders participants in a grid (or, in a vertical grid to the right, if a screen
  /// share is enabled).
  @FreezedUnionValue('default')
  const factory StreamingLayout.defaultLayout({
    /// Specifies how many cameras to include in the grid. Default: 20.
    int? maxCamStreams,
  }) = _DefaultLayout;

  const factory StreamingLayout.singleParticipant({required ParticipantId participant}) = _SingleParticipantLayout;

  const factory StreamingLayout.activeParticipant() = _ActiveParticipantLayout;

  /// A mobile-friendly layout - shows up to two participants in a vertical layout.
  const factory StreamingLayout.portrait({required PortraitVariant variant, int? maxCamStreams}) = _PortraitLayout;

  const factory StreamingLayout.custom({
    required String compositionId,
    Map<String, CompositionParamValue>? compositionParams,
    Map<String, String>? sessionAssets,
  }) = _CustomLayout;

  factory StreamingLayout.fromJson(Map<String, dynamic> json) => _$StreamingLayoutFromJson(json);
}

@freezed
class CompositionParamValue with _$CompositionParamValue {
  const CompositionParamValue._();

  const factory CompositionParamValue.string({required String value}) = _CompositionStringParamValue;

  const factory CompositionParamValue.number({required double value}) = _CompositionNumberParamValue;

  const factory CompositionParamValue.boolean({required bool value}) = _CompositionBooleanParamValue;

  factory CompositionParamValue.fromJson(dynamic json) {
    if (json is String) return CompositionParamValue.string(value: json);
    if (json is num) return CompositionParamValue.number(value: json.toDouble());
    if (json is bool) return CompositionParamValue.boolean(value: json);
    throw ArgumentError('Illegal CompositionParamValue type: ${json.runtimeType}');
  }

  dynamic toJson() => value;
}

@Freezed(unionKey: 'preset')
class LiveStreamEndpoints with _$LiveStreamEndpoints {
  const LiveStreamEndpoints._();

  const factory LiveStreamEndpoints.preconfigured({
    @JsonKey(name: 'preconfiguredEndpoints') required List<Uri> endpoints,
  }) = PreconfiguredLiveStreamEndpoints;

  const factory LiveStreamEndpoints.rtmpUrls({@JsonKey(name: 'rtmpUrls') required List<Uri> urls}) =
      RtmpUrlsLiveStreamEndpoints;

  factory LiveStreamEndpoints.fromJson(Map<String, dynamic> json) => _$LiveStreamEndpointsFromJson(json);
}

@freezed
class RecordingStatus with _$RecordingStatus {
  const factory RecordingStatus({
    /// The id of the recording file.
    RecordingId? recordingId,

    /// The participant who started the recording.
    ParticipantId? startedBy,

    /// Specifies how the participants' videos are laid out in the stream.
    StreamingLayout? layout,

    /// The identifier of the streaming instance.
    required StreamId streamId,
  }) = _RecordingStatus;

  const RecordingStatus._();

  factory RecordingStatus.fromJson(Map<String, dynamic> json) => _$RecordingStatusFromJson(json);
}

@freezed
class LiveStreamStatus with _$LiveStreamStatus {
  const factory LiveStreamStatus({
    /// The participant who started the stream.
    ParticipantId? startedBy,

    /// Specifies how the participants' videos are laid out in the stream.
    StreamingLayout? layout,

    /// The identifier of the live streaming instance.
    required StreamId streamId,
  }) = _LiveStreamStatus;

  factory LiveStreamStatus.fromJson(Map<String, dynamic> json) => _$LiveStreamStatusFromJson(json);
}
